
object NEList {

  def singleton (n:Int):NEList = new Singleton(n)

  def cons (n:Int, L:List):NEList = new Cons(n,L)



  // SINGLETON LIST REPRESENTATION
  //
  private class Singleton (n:Int) extends NEList {

    def isEmpty ():Boolean = false

    def isSingleton ():Boolean = true

    def first ():Int = n

    def rest ():List = List.empty()

    def length ():Int = 1

    def append (M:List):NEList = cons(n,M)

    def find (f:Int):Boolean = (f==n)

    def isEqual (M:List):Boolean = {
      M.length()==1 && M.first()==n
    }

    override def equals (other:Any):Boolean = 
      other match {
	case that:List => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 41

    override def toString ():String = n.toString()
  }


  // CONS LIST REPRESENTATION
  //
  private class Cons (n:Int, L:List) extends NEList {

    def isEmpty ():Boolean = false

    def isSingleton ():Boolean = L.isEmpty()

    def first ():Int = n

    def rest ():List = L

    def length ():Int = 1 + L.length()

    def append (M:List):NEList = cons(n,L.append(M))

    def find (f:Int):Boolean = { (f == n) || L.find(f) }

    def isEqual (M:List):Boolean = 
      { !(M.isEmpty()) && n==M.first() && L.isEqual(M.rest()) }

    override def equals (other:Any):Boolean = 
      other match {
	case that:List => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
	41 + n.hashCode()
      ) + L.hashCode()

    override def toString ():String = n + " " + L.toString()
  }
}



abstract class NEList extends List {

  def isEmpty ():Boolean      
  def isSingleton ():Boolean
  def first ():Int
  def rest ():List
  def length ():Int
  def find (f:Int):Boolean
  def append (M:List):NEList
  def isEqual (M:List):Boolean
}
