
object CPoint {

  def cartesian (x:Double,y:Double,c:Color):CPoint =
    new CarCPoint(x,y,c)


  private class CarCPoint (first:Double, second:Double, 
			   col: Color) extends CPoint {

    def color ():Color = col

    def updateColor (c:Color):CPoint = 
      new CarCPoint(first,second,c)

    def xCoord ():Double = 
      first
    
    def yCoord ():Double = 
      second

    def move (dx:Double,dy:Double):CPoint = 
      new CarCPoint(first+dx, second+dy,col)

    def add (cp:CPoint):CPoint = 
      new CarCPoint(first+cp.xCoord(),second+cp.yCoord(),cp.color())

    def add (p:Point):Point = 
      p match {
        case cp:CPoint => add(cp)
        case _ => Point.cartesian(first+p.xCoord(),second+p.yCoord())
      }

    def rotate (theta:Double):CPoint = 
      new CarCPoint(first * math.cos(theta) - second * math.sin(theta),
                    first * math.sin(theta) + second * math.cos(theta),
          	    col)

    def isEqual (q:CPoint):Boolean = 
      (first==q.xCoord() && second==q.yCoord() && col==q.color())

    def isEqual (q:Point):Boolean = 
      q match {
        case cq:CPoint => isEqual(cq)
	case _ => false
      }

    // CANONICAL

    override def toString ():String = 
      "cart(" + first + "," + second + "," + col + ")"
    
    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }
    
    override def hashCode ():Int = 
      41 * (
	41 * (
	  41 + first.hashCode()
	) + second.hashCode()
      ) + col.hashCode()
  }
}



abstract class CPoint extends Point {

  def color ():Color
  def updateColor (c:Color):CPoint

  def xCoord ():Double
  def yCoord ():Double

  def move (dx:Double,dy:Double):CPoint
  def rotate (t:Double):CPoint

  def add (p:CPoint):CPoint
  def isEqual (p:CPoint):Boolean

  // needed for subtyping
  def add (p:Point):Point
  def isEqual (p:Point):Boolean
}

